/*
 * These hooks are called by the Aragon Buidler plugin during the start task's lifecycle. Use them to perform custom tasks at certain entry points of the development build process, like deploying a token before a proxy is initialized, etc.
 *
 * Link them to the main buidler config file (buidler.config.js) in the `aragon.hooks` property.
 *
 * All hooks receive two parameters:
 * 1) A params object that may contain other objects that pertain to the particular hook.
 * 2) A "bre" or BuidlerRuntimeEnvironment object that contains enviroment objects like web3, Truffle artifacts, etc.
 *
 * Please see AragonConfigHooks, in the plugin's types for further details on these interfaces.
 * https://github.com/aragon/buidler-aragon/blob/develop/src/types.ts#L31
 */
const { hash } = require('eth-ens-namehash')
let token
let oracle
let depositContract

module.exports = {
  // Called before a dao is deployed.
  preDao: async ({ log }, { web3, artifacts }) => {},

  // Called after a dao is deployed.
  postDao: async ({ dao, _experimentalAppInstaller, log }, { web3, artifacts }) => {
    const StETH = artifacts.require('StETH')
    const tokenInstance = await StETH.new()

    const receipt = await dao.newAppInstance(
      hash(`steth.aragonpm.test`), // appId - Unique identifier for each app installed in the DAO; can be any bytes32 string in the tests.
      tokenInstance.address, // appBase - Location of the app's base implementation.
      '0x', // initializePayload - Used to instantiate and initialize the proxy in the same call (if given a non-empty bytes string).
      false // setDefault - Whether the app proxy is the default proxy.
    )
    // Find the deployed proxy address in the tx logs.
    const e = receipt.logs.find((l) => l.event === 'NewAppProxy')
    token = await StETH.at(e.args.proxy)

    const OracleMock = artifacts.require('OracleMock')
    const DepositContractMock = artifacts.require('DepositContractMock')
    depositContract = await DepositContractMock.new()
    oracle = await OracleMock.new()
  },

  // Called after the app's proxy is created, but before it's initialized.
  preInit: async ({ proxy, _experimentalAppInstaller, log }, { web3, artifacts }) => {},

  // Called after the app's proxy is initialized.
  postInit: async ({ proxy, _experimentalAppInstaller, log }, { web3, artifacts }) => {},

  // Called when the start task needs to know the app proxy's init parameters.
  // Must return an array with the proxy's init parameters.
  getInitParams: async ({ log }, { web3, artifacts }) => {
    return [token.address, depositContract.address, oracle.address, 10]
  },

  // Called after the app's proxy is updated with a new implementation.
  postUpdate: async ({ proxy, log }, { web3, artifacts }) => {}
}
